@@ -6,7 +6,26 @@
 (*         *       GNU Lesser General Public License Version 2.1        *)
 (************************************************************************)
 
-open Compat (* necessary for camlp4 *)
+open GramCompat (* necessary for camlp4 *)
+
+type argument_type =
+| ListArgType of argument_type
+| OptArgType of argument_type
+| PairArgType of argument_type * argument_type
+| ExtraArgType of string
+
+type user_symbol =
+| Ulist1 of user_symbol
+| Ulist1sep of user_symbol * string
+| Ulist0 of user_symbol
+| Ulist0sep of user_symbol * string
+| Uopt of user_symbol
+| Uentry of string
+| Uentryl of string * int
+
+type extend_token =
+| ExtTerminal of string
+| ExtNonTerminal of user_symbol * string
 
 val mlexpr_of_list :  ('a -> MLast.expr) -> 'a list -> MLast.expr
 
@@ -14,14 +33,6 @@ val mlexpr_of_pair :
   ('a -> MLast.expr) -> ('b -> MLast.expr)
     -> 'a * 'b -> MLast.expr
 
-val mlexpr_of_triple :
-  ('a -> MLast.expr) -> ('b -> MLast.expr) -> ('c -> MLast.expr)
-    -> 'a * 'b * 'c -> MLast.expr
-
-val mlexpr_of_quadruple :
-  ('a -> MLast.expr) -> ('b -> MLast.expr) ->
-    ('c -> MLast.expr) -> ('d -> MLast.expr) -> 'a * 'b * 'c * 'd -> MLast.expr
-
 val mlexpr_of_bool : bool -> MLast.expr
 
 val mlexpr_of_int : int -> MLast.expr
@@ -30,4 +41,8 @@ val mlexpr_of_string : string -> MLast.expr
 
 val mlexpr_of_option : ('a -> MLast.expr) -> 'a option -> MLast.expr
 
-val mlexpr_of_prod_entry_key : Pcoq.prod_entry_key -> MLast.expr
+val mlexpr_of_prod_entry_key : (string -> MLast.expr) -> user_symbol -> MLast.expr
+
+val type_of_user_symbol : user_symbol -> argument_type
+
+val parse_user_entry : string -> string -> user_symbol
