@@ -1,6 +1,6 @@
 open Pp
 open Glob_term
-open Errors
+open CErrors
 open Util
 open Names
 open Decl_kinds
